import iro
import argparse


def main():
    parser = argparse.ArgumentParser(description="Host an Iro server")

    parser.add_argument(
        "-a", "--addr", help="address to host on", type=str, required=False
    )
    parser.add_argument("-p", "--port", help="port to listen on", type=int)
    parser.add_argument("-f", "--filter", help="sniffing filter to use", type=str)
    parser.add_argument("-c", "--cert", help="SSL cert path", type=str)
    parser.add_argument("-k", "--key", help="SSL key path", type=str)
    args = parser.parse_args()

    iro.start(args.addr, args.port, args.filter, args.cert, args.key)



from toori import TooriClient
import argparse


def main():
    parser = argparse.ArgumentParser(description="Connect to an Iro server")

    mode_group = parser.add_mutually_exclusive_group(required=True)
    mode_group.add_argument("-g", "--gui", action="store_true", help="Toori GUI")
    mode_group.add_argument(
        "-a", "--addr", help="address of server to connect to", type=str, required=False
    )

    parser.add_argument("-p", "--port", help="server port to connect to", type=int)
    parser.add_argument("-k", "--key", help="key to encrypt packets", type=str)
    parser.add_argument("-f", "--filter", help="sniffing filter to use", type=str)
    parser.add_argument(
        "-t",
        "--transports",
        help="socketIO transport method",
        type=str,
        choices=["websocket", "polling", "auto"],
    )

    args = parser.parse_args()

    if args.gui:
        from toori.gui import gui_main

        gui_main()
    else:
        cli = TooriClient(args.addr, args.port, args.filter, args.transports, args.key)
        cli.start()


if __name__ == "__main__":
    from toori.gui import gui_main

    gui_main()



"""
色
"""

from engineio.payload import Payload
import socketio
from aiohttp import web
import asyncio
import ssl
from scapy.all import (
    IP,
    TCP,
    AsyncSniffer,
    conf,
    get_if_addr,
    UDP,
)

# Can be generic because most packets are filtered on the client side
sniff_filter = "tcp and (src port 80 or src port 443)"

# Prevents crashing when overloaded
Payload.max_decode_packets = 250000

# Local IP of the server
LOCAL_IP = get_if_addr(conf.iface)

# Set to whitelist dst IPs initiated by client
dest_whitelist = set()

# L3 socket object for Scapy to reuse
scapy_l3_socket = conf.L3socket()

# Ping timeout prevents client from disconnecting when there is a delay
sio = socketio.AsyncServer(
    async_mode="aiohttp", ping_timeout=9999, serializer="default"
)

app = web.Application()
sio.attach(app)


async def handleInboundPacket(pkt):
    try:
        # Get packet from IP layer onwards and send to client
        packet_inbound = pkt[IP]
    except IndexError:
        return

    # Skip packet if it is not a reply to the client
    if packet_inbound[IP].src not in dest_whitelist:
        return

    # packet_inbound.show()

    # Send packet to client
    await sio.emit(
        "ret",
        bytes(packet_inbound),
    )


# Sniff for returning inbound packets in the background
async def background_task():
    sniffer = AsyncSniffer(
        filter=sniff_filter,
        store=False,
        prn=lambda pkt: asyncio.run(handleInboundPacket(pkt)),
    )
    print(f"Sniffing with filter ({sniff_filter})")
    sniffer.start()


# Listen for messages on "out" event
@sio.on("out")
async def handleOutboundPacket(sid, data):
    packet_outbound = IP(data)
    # packet_outbound.show()

    # Change packet source IP to server's local IP
    packet_outbound[IP].src = LOCAL_IP

    # Whitelist st IP to allow returning packet
    dest_whitelist.add(packet_outbound[IP].dst)

    # Force recalculaton of layer checksums
    if packet_outbound.haslayer(IP):
        del packet_outbound[IP].chksum
    if packet_outbound.haslayer(TCP):
        del packet_outbound[TCP].chksum
    if packet_outbound.haslayer(UDP):
        del packet_outbound[UDP].chksum

    scapy_l3_socket.send(packet_outbound)


# On client connect
@sio.event
async def connect(sid, environ):
    print(f"Client {sid} connected")

    # Clear whitelist
    global dest_whitelist
    dest_whitelist.clear()


# On client disconnect
@sio.event
def disconnect(sid):
    print(f"Client {sid} disconnected")

    # Clear whitelist
    global dest_whitelist
    dest_whitelist.clear()


# Start background task and initialize
async def init_app():
    sio.start_background_task(background_task)
    return app


def start(host, port, filter=None, ssl_cert_path=None, ssl_key_path=None):
    ssl_context = None

    # "/etc/letsencrypt/live//fullchain.pem"
    # "/etc/letsencrypt/live//privkey.pem"
    if ssl_cert_path and ssl_key_path:
        ssl_context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
        ssl_context.load_cert_chain(ssl_cert_path, ssl_key_path)

    if not host:
        host = "0.0.0.0"
    if not port:
        port = 443

    if filter:
        global sniff_filter
        sniff_filter = filter

    web.run_app(init_app(), host=host, port=port, ssl_context=ssl_context)


if __name__ == "__main__":
    # Debug
    start("localhost", 8080)



"""
通り
"""

from _toori import *
from .constants import *

import socketio
import socket

from cryptography.fernet import Fernet
import base64

import urllib3

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

from engineio.payload import Payload

# Increase the packet buffer
Payload.max_decode_packets = 2500000


class TooriClient:
    # Attempts WebSocket first, else fallback to polling
    def __init__(
        self,
        server_url: str,
        server_port: int,
        pkt_filter: str,
        transport: str,
        pkt_key: str,
        loggerSignal=None,
        connectedSignal=None,
    ):
        # Set default parameters if not given
        if not server_url:
            server_url = URL_DEFAULT
        if not server_port:
            server_port = PORT_DEFAULT
        if not pkt_filter:
            pkt_filter = FILTER_DEFAULT
        if not transport:
            transport = TRANSPORT_DEFAULT
        if not pkt_key:
            pkt_key = PKT_KEY_DEFAULT

        self.connectedSignal = connectedSignal

        # Set logging function to signal instead of print if GUI is used
        self.log = print
        if loggerSignal:
            self.loggerSignal = loggerSignal
            self.log = self._logger

        self.fernet = Fernet(base64.urlsafe_b64encode(pkt_key.encode()))

        # Transport must be None for auto
        self.log(f"Using {transport} transport")
        if transport == TRANSPORT_DEFAULT:
            transport = None

        # Intercept only outbound and non loopback traffic
        # And allow traffic to the Iro server
        server_ip = self.get_ip(server_url)
        pkt_filter = (
            f"outbound && !loopback && ip.DstAddr != {server_ip} && ({pkt_filter})"
        )
        self.log(f"Using filter ({pkt_filter})")

        # Add SSL if not specified
        if not server_url.startswith("http"):
            if server_port == 443:
                server_url = f"https://{server_url}"
            else:
                server_url = f"http://{server_url}"

        self.local_ip = self.get_local_ip()

        self.sio = socketio.Client(ssl_verify=False)

        self.server_url = server_url
        self.server_port = server_port
        self.pkt_filter = pkt_filter
        self.transport = transport

    def get_ip(self, url):
        return socket.gethostbyname(url.split("//")[-1])

    def get_local_ip(self):
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        s.connect(("8.8.8.8", 80))
        return s.getsockname()[0]

    def _logger(self, *args):
        self.loggerSignal.emit(" ".join(*args))

    def uninit_WinDivert(self):
        uninit_outbound()
        uninit_inbound()
        self.log("Toori unestablished")

    def disconnect_sio(self):
        self.sio.disconnect()
        self.log("Disconnected from server")

    def start_inbound_handler(self):
        @self.sio.on("ret")
        def injector(data):
            if not self.sio.connected:
                return
            inject_once(self.fernet.decrypt(data))

    def connect(self):
        self.log(f"Connecting to {self.server_url}:{self.server_port}")
        try:
            self.sio.connect(
                f"{self.server_url}:{self.server_port}", transports=self.transport
            )
        except socketio.exceptions.ConnectionError:
            self.log("Could not connect")
            return -1

        self.log(f"Connected to server")

    def start(self):
        if self.connect() == -1:
            return

        try:
            # Initialize WinDivert injector with address to replace
            init_inbound(self.local_ip)
        except (TooriException, WinDivertException) as e:
            self.log(str(e))
            self.disconnect_sio()
            return

        try:
            # Initialize WinDivert sniffer with filter
            init_outbound(self.pkt_filter)
        except (TooriException, WinDivertException) as e:
            self.log(str(e))
            self.disconnect_sio()
            uninit_inbound()
            return

        self.start_inbound_handler()

        if self.connectedSignal:
            self.connectedSignal.emit(1)

        self.log(f"Toori established")

        while True:
            try:
                self.sio.emit("out", self.fernet.encrypt(recv_once()))
            except socketio.exceptions.BadNamespaceError:
                continue